// Copyright (C) 2021  Simon Ruderich
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

package main_test

import (
	"fmt"
	"net"
	"os"
	"os/exec"
	"regexp"
	"runtime"
	"strings"
	"testing"
	"time"

	"ruderich.org/simon/safcm/testutil"
)

func TestSyncSshEndToEnd(t *testing.T) {
	cwd, err := os.Getwd()
	if err != nil {
		t.Fatal(err)
	}
	defer os.Chdir(cwd)

	var suffix string
	// Needs different options in sshd_config
	if runtime.GOOS == "openbsd" {
		suffix = ".openbsd"
	}

	sshDir := cwd + "/testdata/ssh"
	sshCmd := exec.Command("/usr/sbin/sshd",
		"-D", // stay in foreground
		"-e", // write messages to stderr instead of syslog
		"-f", sshDir+"/sshd/sshd_config"+suffix,
		"-h", sshDir+"/sshd/ssh_host_key",
		"-o", "AuthorizedKeysFile="+sshDir+"/ssh/authorized_keys",
	)
	sshCmd.Stderr = os.Stderr
	err = sshCmd.Start()
	if err != nil {
		t.Fatal(err)
	}
	defer sshCmd.Process.Kill()

	// Wait until SSH server is ready (up to 30 seconds)
	for i := 0; i < 30; i++ {
		conn, err := net.Dial("tcp", "127.0.0.1:29327")
		if err == nil {
			conn.Close()
			break
		}
		time.Sleep(time.Second)
	}

	err = os.Chdir(sshDir + "/project")
	if err != nil {
		t.Fatal(err)
	}

	tests := []struct {
		name   string
		remove bool
		args   []string
		exp    string
		expErr error
	}{

		{
			"no settings",
			true,
			[]string{"no-settings.example.org"},
			`<LOG>[info]    [no-settings.example.org] remote helper upload in progress
<LOG>[info]    [no-settings.example.org] no changes
`,
			nil,
		},
		{
			"no settings (no helper upload)",
			false,
			[]string{"no-settings.example.org"},
			`<LOG>[info]    [no-settings.example.org] no changes
`,
			nil,
		},
		{
			"no settings (error)",
			true,
			[]string{"-log", "error", "no-settings.example.org"},
			``,
			nil,
		},
		{
			"no settings (verbose)",
			true,
			[]string{"-log", "verbose", "no-settings.example.org"},
			`<LOG>[info]    [no-settings.example.org] remote helper upload in progress
<LOG>[verbose] [no-settings.example.org] host groups: all <DET> <DET> no-settings.example.org
<LOG>[verbose] [no-settings.example.org] host group priorities (desc. order): no-settings.example.org
<LOG>[info]    [no-settings.example.org] no changes
`,
			nil,
		},
		{
			"no settings (debug2)",
			true,
			[]string{"-log", "debug2", "no-settings.example.org"},
			`<LOG>[info]    [no-settings.example.org] remote helper upload in progress
<LOG>[verbose] [no-settings.example.org] host groups: all <DET> <DET> no-settings.example.org
<LOG>[verbose] [no-settings.example.org] host group priorities (desc. order): no-settings.example.org
<LOG>[info]    [no-settings.example.org] no changes
`,
			nil,
		},
	}

	remotePath := fmt.Sprintf("/tmp/safcm-remote-%d", os.Getuid())

	logRegexp := regexp.MustCompile(`^\d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2} `)
	detectedRegexp := regexp.MustCompile(`detected_\S+`)

	for _, tc := range tests {
		t.Run(tc.name, func(t *testing.T) {
			if tc.remove {
				os.Remove(remotePath)
			}

			args := append([]string{"sync",
				"-sshconfig", sshDir + "/ssh/ssh_config",
			}, tc.args...)
			cmd := exec.Command("../../../../../safcm", args...)
			out, err := cmd.CombinedOutput()

			var tmp []string
			for _, x := range strings.Split(string(out), "\n") {
				// Strip parts which change on each run (LOG)
				// or depending on the system (DET)
				x = logRegexp.ReplaceAllString(x, "<LOG>")
				x = detectedRegexp.ReplaceAllString(x, "<DET>")
				tmp = append(tmp, x)
			}
			res := strings.Join(tmp, "\n")

			testutil.AssertEqual(t, "res", res, tc.exp)
			testutil.AssertErrorEqual(t, "err", err, tc.expErr)
		})
	}

	os.Remove(remotePath)
}
